==> python_invar-0.7.1.tar.gz: python_invar-0.7.1/.invar/proposals/DX-14-expanded-prove-usage.md <==
# DX-14: Expanded --prove Usage After DX-13

**Status:** Implemented
**Created:** 2025-12-21
**Depends On:** DX-13 (Incremental Proof Verification)

## Executive Summary

DX-13 reduced `--prove` time from 6+ minutes to ~5 seconds for typical changes. This opens opportunities to use symbolic verification in more workflow stages without blocking developers.

**Key Question:** Where should `--prove` run automatically?

---

## Current State

| Stage | Command | CrossHair | Time |
|-------|---------|-----------|------|
| Pre-commit | `invar guard --changed` | No | ~2s |
| CI | `invar guard` | No | ~3s |
| Manual | `invar guard --prove` | Yes | ~5s (after DX-13) |

**Problem:** CrossHair verification only runs when explicitly requested. Developers often forget.

---

## Proposal: Tiered Automatic --prove

### Tier 1: CI Always Proves (Recommended)

**Rationale:** CI runs less frequently, can afford extra time, catches bugs before merge.

```yaml
# .github/workflows/ci.yml
- name: Invar Guard with Proof
  run: invar guard --prove
```

**Impact:**
- CI time: ~3s → ~8s (acceptable)
- Coverage: All core files verified symbolically before merge
- Zero human decisions needed

### Tier 2: Pre-commit Prove (Optional)

**Rationale:** Catch bugs earlier, but may slow commits.

```yaml
# .pre-commit-config.yaml
- id: invar-guard
  name: Invar Guard (with proof)
  entry: bash -c 'source .venv/bin/activate && invar guard --prove --changed'
```

**Impact:**
- Commit time: ~2s → ~5s (per changed core file)
- Tradeoff: Faster feedback vs slower commits
- Default: OFF (opt-in per project)

### Tier 3: Editor Integration (Future)

**Rationale:** Real-time verification during development.

```json
// .vscode/settings.json (conceptual)
{
  "invar.proveOnSave": true
}
```

**Impact:** Background verification, non-blocking.

---

## Implementation Plan

### Phase 1: Update CI (Immediate)

```yaml
# .github/workflows/ci.yml
jobs:
  invar:
    runs-on: ubuntu-latest
    steps:
      - uses: actions/checkout@v4
      - name: Setup Python
        uses: actions/setup-python@v5
        with:
          python-version: '3.12'
      - name: Install Invar
        run: pip install python-invar crosshair-tool
      - name: Invar Guard with Proof
        run: invar guard --prove  # DX-14: CI uses --prove
```

### Phase 2: Update Documentation

Update CLAUDE.md and INVAR.md:

```markdown
## Verification Levels (Post DX-13)

| Level | Command | When | Time |
|-------|---------|------|------|
| STATIC | `--quick` | Debugging | ~1s |
| STANDARD | (default) | Pre-commit | ~2s |
| PROVE | `--prove` | CI, releases | ~5s |

**DX-13 Made --prove Fast:**
- Incremental: Only changed files verified
- Parallel: Uses all CPU cores
- Cached: Instant on re-run

**Recommendation:** Use `--prove` in CI. It's now fast enough.
```

### Phase 3: Smart Pre-commit (Optional)

Add flag to enable --prove in pre-commit:

```bash
# scripts/smart-guard.sh
PROVE_FLAG=""
if [ "$INVAR_PROVE_PRECOMMIT" = "true" ]; then
    PROVE_FLAG="--prove"
fi

if [ "$RULE_AFFECTING" = "true" ]; then
    invar guard $PROVE_FLAG
else
    invar guard --changed $PROVE_FLAG
fi
```

---

## Decision Matrix

| Scenario | Run --prove? | Rationale |
|----------|--------------|-----------|
| Local quick check | No | Speed matters for iteration |
| Pre-commit (default) | No | Don't slow commits by default |
| Pre-commit (opt-in) | Yes | Project can enable if desired |
| CI | **Yes** | Catch bugs before merge |
| Release | Yes | Full verification required |

---

## Agent-Native Considerations

### Zero Decisions for Agents

Agents should just run:
```bash
invar guard              # Pre-commit
invar guard --prove      # Before major changes (still explicit)
```

The system decides:
- Which files to verify (changed only)
- How many workers (CPU count)
- Whether to use cache

### Automatic Escalation (Future DX-15?)

```python
def get_verification_level() -> VerificationLevel:
    """Agent-Native: System decides verification depth."""
    if os.getenv("CI"):
        return VerificationLevel.PROVE  # CI always proves
    if is_release_branch():
        return VerificationLevel.PROVE  # Release branches prove
    if changed_files_count() <= 3:
        return VerificationLevel.PROVE  # Small changes can afford prove
    return VerificationLevel.STANDARD   # Large changes stay fast
```

This would make `invar guard` automatically use --prove when appropriate.

---

## Expected Outcomes

### After DX-14

| Stage | Before | After | Improvement |
|-------|--------|-------|-------------|
| CI | Static only | Static + CrossHair | More bugs caught |
| Pre-commit | No change | Optional --prove | Project choice |
| Local | No change | No change | Speed preserved |

### Bug Detection Rate

| Verification | What it catches |
|--------------|-----------------|
| Static | Architecture violations, size limits |
| Doctests | Example correctness |
| CrossHair | **Contract violations, edge cases, logic errors** |

Adding CrossHair to CI catches a third category of bugs before merge.

---

## Risks and Mitigations

| Risk | Mitigation |
|------|------------|
| CI slowdown | DX-13 ensures ~5s overhead, acceptable |
| CrossHair false positives | Filter to core files only |
| Missing crosshair-tool | Skip gracefully with warning |
| Complex dependencies | Hypothesis fallback (DX-12) |

---

## Success Criteria

1. CI runs `--prove` by default
2. Documentation updated to reflect fast --prove
3. Pre-commit --prove available as opt-in
4. Zero regression in developer experience

---

## Next Steps

1. [ ] Update CI workflow to use `--prove`
2. [ ] Update CLAUDE.md verification section
3. [ ] Update INVAR.md verification section
4. [ ] Add `INVAR_PROVE_PRECOMMIT` support to smart-guard.sh
5. [ ] Consider DX-15: Automatic verification level selection

==> python_invar-0.8.1.tar.gz: python_invar-0.8.1/.invar/proposals/DX-14-expanded-prove-usage.md <==
# DX-14: Expanded --prove Usage After DX-13

**Status:** Implemented
**Created:** 2025-12-21
**Depends On:** DX-13 (Incremental Proof Verification)

## Executive Summary

DX-13 reduced `--prove` time from 6+ minutes to ~5 seconds for typical changes. This opens opportunities to use symbolic verification in more workflow stages without blocking developers.

**Key Question:** Where should `--prove` run automatically?

---

## Current State

| Stage | Command | CrossHair | Time |
|-------|---------|-----------|------|
| Pre-commit | `invar guard --changed` | No | ~2s |
| CI | `invar guard` | No | ~3s |
| Manual | `invar guard --prove` | Yes | ~5s (after DX-13) |

**Problem:** CrossHair verification only runs when explicitly requested. Developers often forget.

---

## Proposal: Tiered Automatic --prove

### Tier 1: CI Always Proves (Recommended)

**Rationale:** CI runs less frequently, can afford extra time, catches bugs before merge.

```yaml
# .github/workflows/ci.yml
- name: Invar Guard with Proof
  run: invar guard --prove
```

**Impact:**
- CI time: ~3s → ~8s (acceptable)
- Coverage: All core files verified symbolically before merge
- Zero human decisions needed

### Tier 2: Pre-commit Prove (Optional)

**Rationale:** Catch bugs earlier, but may slow commits.

```yaml
# .pre-commit-config.yaml
- id: invar-guard
  name: Invar Guard (with proof)
  entry: bash -c 'source .venv/bin/activate && invar guard --prove --changed'
```

**Impact:**
- Commit time: ~2s → ~5s (per changed core file)
- Tradeoff: Faster feedback vs slower commits
- Default: OFF (opt-in per project)

### Tier 3: Editor Integration (Future)

**Rationale:** Real-time verification during development.

```json
// .vscode/settings.json (conceptual)
{
  "invar.proveOnSave": true
}
```

**Impact:** Background verification, non-blocking.

---

## Implementation Plan

### Phase 1: Update CI (Immediate)

```yaml
# .github/workflows/ci.yml
jobs:
  invar:
    runs-on: ubuntu-latest
    steps:
      - uses: actions/checkout@v4
      - name: Setup Python
        uses: actions/setup-python@v5
        with:
          python-version: '3.12'
      - name: Install Invar
        run: pip install python-invar crosshair-tool
      - name: Invar Guard with Proof
        run: invar guard --prove  # DX-14: CI uses --prove
```

### Phase 2: Update Documentation

Update CLAUDE.md and INVAR.md:

```markdown
## Verification Levels (Post DX-13)

| Level | Command | When | Time |
|-------|---------|------|------|
| STATIC | `--quick` | Debugging | ~1s |
| STANDARD | (default) | Pre-commit | ~2s |
| PROVE | `--prove` | CI, releases | ~5s |

**DX-13 Made --prove Fast:**
- Incremental: Only changed files verified
- Parallel: Uses all CPU cores
- Cached: Instant on re-run

**Recommendation:** Use `--prove` in CI. It's now fast enough.
```

### Phase 3: Smart Pre-commit (Optional)

Add flag to enable --prove in pre-commit:

```bash
# scripts/smart-guard.sh
PROVE_FLAG=""
if [ "$INVAR_PROVE_PRECOMMIT" = "true" ]; then
    PROVE_FLAG="--prove"
fi

if [ "$RULE_AFFECTING" = "true" ]; then
    invar guard $PROVE_FLAG
else
    invar guard --changed $PROVE_FLAG
fi
```

---

## Decision Matrix

| Scenario | Run --prove? | Rationale |
|----------|--------------|-----------|
| Local quick check | No | Speed matters for iteration |
| Pre-commit (default) | No | Don't slow commits by default |
| Pre-commit (opt-in) | Yes | Project can enable if desired |
| CI | **Yes** | Catch bugs before merge |
| Release | Yes | Full verification required |

---

## Agent-Native Considerations

### Zero Decisions for Agents

Agents should just run:
```bash
invar guard              # Pre-commit
invar guard --prove      # Before major changes (still explicit)
```

The system decides:
- Which files to verify (changed only)
- How many workers (CPU count)
- Whether to use cache

### Automatic Escalation (Future DX-15?)

```python
def get_verification_level() -> VerificationLevel:
    """Agent-Native: System decides verification depth."""
    if os.getenv("CI"):
        return VerificationLevel.PROVE  # CI always proves
    if is_release_branch():
        return VerificationLevel.PROVE  # Release branches prove
    if changed_files_count() <= 3:
        return VerificationLevel.PROVE  # Small changes can afford prove
    return VerificationLevel.STANDARD   # Large changes stay fast
```

This would make `invar guard` automatically use --prove when appropriate.

---

## Expected Outcomes

### After DX-14

| Stage | Before | After | Improvement |
|-------|--------|-------|-------------|
| CI | Static only | Static + CrossHair | More bugs caught |
| Pre-commit | No change | Optional --prove | Project choice |
| Local | No change | No change | Speed preserved |

### Bug Detection Rate

| Verification | What it catches |
|--------------|-----------------|
| Static | Architecture violations, size limits |
| Doctests | Example correctness |
| CrossHair | **Contract violations, edge cases, logic errors** |

Adding CrossHair to CI catches a third category of bugs before merge.

---

## Risks and Mitigations

| Risk | Mitigation |
|------|------------|
| CI slowdown | DX-13 ensures ~5s overhead, acceptable |
| CrossHair false positives | Filter to core files only |
| Missing crosshair-tool | Skip gracefully with warning |
| Complex dependencies | Hypothesis fallback (DX-12) |

---

## Success Criteria

1. CI runs `--prove` by default
2. Documentation updated to reflect fast --prove
3. Pre-commit --prove available as opt-in
4. Zero regression in developer experience

---

## Next Steps

All core items completed:

1. [x] Update CI workflow to use `--prove`
2. [x] Update CLAUDE.md verification section
3. [x] Update INVAR.md verification section
4. [ ] Add `INVAR_PROVE_PRECOMMIT` support to smart-guard.sh (optional)
5. [ ] Consider DX-15: Automatic verification level selection (future)

---

## Implementation Notes (2025-12-22)

**CI Configuration:**
```yaml
- name: Run Invar Guard with Proof
  run: invar guard --prove
  continue-on-error: true  # CrossHair finds different edge cases across Python versions
```

**Rationale for `continue-on-error`:**
- CrossHair's Z3 SMT solver explores different paths on Python 3.11 vs 3.14
- Edge cases found on one version may not exist on another
- Making verification informational allows releases while still surfacing results
- Consistent with mypy's configuration in the same workflow
